import tkinter as tk
from TuneCoach.gui.constants import *

class Graph(tk.Canvas):
    def clear_plot(self):
        self.delete("points")
    
    def update_plot(self, score_history):
        # self.session.data.update_score_history()
        numScores = len(score_history)
        self.cached_scores = score_history
        self.clear_plot()
        if numScores != 0:
            for i in range(0, numScores-1):
                yCoord = int((1 - (score_history[i])/100) * (self.canvasHeight-40)+10)
                xCoord = int((i)/(self.scores_to_display-1) * (self.canvasWidth - 55)) + 40
                nextY = int((1-(score_history[i+1])/100)*(self.canvasHeight-40)+10)
                nextX = int((i+1)/(self.scores_to_display-1)*(self.canvasWidth-55))+40
                self.create_circle(xCoord, yCoord, 4)
                self.create_line(xCoord, yCoord, nextX, nextY, tag="points", fill=Colors.graph_foreground)
            yCoord = int((1-(score_history[numScores-1])/100) * (self.canvasHeight-40)+10)
            xCoord = int((numScores-1)/(self.scores_to_display-1)*(self.canvasWidth-55))+40
            self.create_circle(xCoord, yCoord, 4)

    def create_circle(self, x, y, r):
        x0 = x - r
        y0 = y - r
        x1 = x + r
        y1 = y + r
        return self.create_oval(x0, y0, x1, y1, tag="points", fill=Colors.graph_foreground, outline="")

    def __init__(self, frame, **kwargs):
        super().__init__(frame, kwargs, bg=Colors.graph_background, bd=0, highlightthickness=0)
        self.frame = frame
        self.canvasHeight = self.winfo_height()
        self.canvasWidth = self.winfo_width()
        self.cached_scores = []
        self.scores_to_display = 15
        self.pack()
        self.draw(None)
        self.bind("<Configure>", self.draw)

    def draw(self, event):
        self.delete("all")
        self.canvasHeight = self.winfo_height()
        self.canvasWidth = self.winfo_width()

        self.create_line(40, 10, 40, self.canvasHeight-30, fill=Colors.graph_foreground)
        self.create_line(40, self.canvasHeight-30, self.canvasWidth-15, self.canvasHeight-30, fill=Colors.graph_foreground)
        self.create_text(self.canvasWidth/2, self.canvasHeight-15, text="Total Score over Time", fill=Colors.graph_foreground)

        self.create_text(20, 10, text="100%", fill = Colors.graph_foreground)
        self.create_text(20, int((self.canvasHeight-40)*.2+10), text="80%", fill=Colors.graph_foreground)
        self.create_text(20, int((self.canvasHeight-40)*.4+10), text="60%", fill=Colors.graph_foreground)
        self.create_text(20, int((self.canvasHeight-40)*.6+10), text="40%", fill=Colors.graph_foreground)
        self.create_text(20, int((self.canvasHeight-40)*.8+10), text="20%", fill=Colors.graph_foreground)

        self.create_text(20, int((self.canvasHeight-40)+10), text="0", fill=Colors.graph_foreground)

        self.create_line(40, 10, 44, 10, fill=Colors.graph_foreground)
        self.create_line(40, int((self.canvasHeight-40)*.2+20), 34, int((self.canvasHeight-40)*.2+20), fill=Colors.graph_foreground)
        self.create_line(40, int((self.canvasHeight-40)*.4+20), 34, int((self.canvasHeight-40)*.4+20), fill=Colors.graph_foreground)
        self.create_line(40, int((self.canvasHeight-40)*.6+20), 34, int((self.canvasHeight-40)*.6+20), fill=Colors.graph_foreground)
        self.create_line(40, int((self.canvasHeight-40)*.8+20), 34, int((self.canvasHeight-40)*.8+20), fill=Colors.graph_foreground)

        self.update_plot(self.cached_scores) # So that redrawing while data isn't coming in doesn't clear existing points



from TuneCoach.python_bridge.SessionData import load_from_file, save_to_file
from pathlib import Path


def save_session(session):
    if session.path is not None:
        save_to_file(session.data, session.path)
    else:
        raise AttributeError('Cannot save a session without a path')


def load_session(path):
    data = load_from_file(path)
    if data is None:
        return None
    return Session(data, path)


class Session:
    def __init__(self, data, path = None):
        self.data = data
        self.path = path
        if self.path is None:
            self.name = "New Session"
        else:
            self.name = Path(self.path).stem
    
    # rather than edit the path and data directly, just create a new
    # Session object with the new parameters. This is cheap because
    # the data is stored separately
    def with_path(self, path):
        return Session(self.data, path)



import tkinter as tk
import tkinter.ttk as ttk
import TuneCoach.gui as gui
from TuneCoach.gui.Piano import Piano
from TuneCoach.gui.constants import *
from TuneCoach.python_bridge.pitch_utilities import *


# Different classes for pop-up windows.
class SessionHistory:
    def create_note(self, x, y, r, canvas, fillColor):  # center coordinates, radius
        x0 = x - r
        y0 = y - r
        x1 = x + r
        y1 = y + r
        return canvas.create_rectangle(x0, y0, x1, y1, fill=fillColor, outline="")

    def __init__(self, mainWindow, workingFrame):
        self.mainWindow = mainWindow
        self.frame = workingFrame
        self.canvas = tk.Canvas(workingFrame, bg=Colors.piano_track, bd=0, highlightthickness=0)
        self.canvas.pack(expand=True, fill=tk.BOTH)

        self.display_size = 64
        self.current_pos = 0

        self.scrollbar = tk.Scrollbar(workingFrame, orient=tk.HORIZONTAL)
        self.scrollbar.pack(side='bottom', fill='x')
        self.scrollbar.config(command=self.scroll)
        self.scrollbar.config(bg=Colors.aux, activebackground=Colors.tuner_needle, troughcolor=Colors.scroll_trough)
        self.scrollbar_width = 1
        self.scrollbar.set(1 - self.scrollbar_width, 1)

        self.aspect_ratio = 580/820
        self.piano = Piano(self.canvas, width=50, height=90)
        self.piano.pack(side='bottom', expand=True, fill='y', anchor='w')

        self.circle_size = 5
        self.circle_start = 0
        self.circle_list = [None] * self.display_size

        self.frame.bind("<Configure>", self.setup)
        self.canvas.bind("<Configure>", self.setup)
        self.setup(None)
        self.pause_scrollbar()

    @property
    def buffer(self):
        return self.mainWindow.controller.session.data.note_history
    
    def scroll(self, *args):
        if args[0] == 'update_width':
            self.scrollbar_width = args[1]
            self.scrollbar.set(1 - self.scrollbar_width, 1)
        if args[0] == 'moveto':
            offset = max(0, min(float(args[1]), 1 - self.scrollbar_width))
            self.scrollbar.set(offset, offset + self.scrollbar_width)
            self.display_notes(int(len(self.buffer) * offset))
        elif args[0] == 'scroll' and len(self.buffer) > 0:
            amount = int(args[1])
            if args[2] == 'pages':
                self.display_notes(self.current_pos + int(0.5*self.display_size*amount))
            elif args[2] == 'units':
                self.display_notes(self.current_pos + amount)
            offset = self.current_pos/len(self.buffer)
            self.scrollbar.set(offset, offset + self.scrollbar_width)
        
    def setup(self, event): # TODO: fix bug where resizing window removes current data from display
        self.clear()
        self.canvas.delete("all")

        width = self.frame.winfo_width()
        height = self.canvas.winfo_height() # Don't count the scrollbar

        piano_width = height*self.aspect_ratio
        self.piano.configure(width=piano_width)
            
        available_width = width - piano_width
        self.circle_start = piano_width
        self.circle_size = 0.5*available_width/self.display_size
        
        self.noteDict = {
            11: height / 14,
            10: height / 7,
            9: height / 14 * 3,
            8: height / 7 * 2,
            7: height / 14 * 5,
            6: height / 7 * 3,
            5: height / 14 * 7,
            4: height / 14 * 9,
            3: height / 7 * 5,
            2: height / 14 * 11,
            1: height / 7 * 6,
            0: height / 14 * 13
        }

        for note in self.noteDict:
            self.canvas.create_line(piano_width, self.noteDict[note], width, self.noteDict[note], width=0.5)

        self.display_notes(self.current_pos) # redraw the notes the user was currently looking at

    def update(self, data):
        if len(data.note_history) > 0:
            self.scroll('update_width', 1 / max(1, len(self.buffer) / self.display_size))
            pitch_errors = [(100.0 * data._in_tune_count[i]) / (data._pitch_count[i] if data._pitch_count[i] != 0 else 1) for i in range(0,12)]
            self.piano.set_scores(pitch_errors, data.key_signature)
            self.display_recent_notes()

    def display_recent_notes(self):
        self.display_notes(max(0, len(self.buffer) - self.display_size))

    def display_notes(self, pos):
        if len(self.buffer) <= self.display_size:
            pos = 0
        else:
            pos = min(max(0, pos), len(self.buffer) - self.display_size)
        self.current_pos = pos
        notes = self.buffer[pos : pos + self.display_size]
        for i, note in enumerate(notes):
            color = Colors.red
            if abs(note.cents) <= self.mainWindow.controller.threshold:
                color = Colors.green
            elif abs(note.cents) <= self.mainWindow.controller.yellow_threshold:
                color = Colors.yellow

            circle = self.circle_list[i]
            x = self.circle_start + self.circle_size + 2*self.circle_size*i
            y = self.noteDict[note.pitch_class]
            if circle is None:
                c = self.create_note(x, y, self.circle_size, self.canvas, color)
                self.circle_list[i] = c
            else:
                self.canvas.coords(circle, x - self.circle_size, y - self.circle_size, x + self.circle_size, y + self.circle_size)
                self.canvas.itemconfig(circle, fill=color)

    def clear(self):
        for circle in self.circle_list:
            self.canvas.delete(circle)
        self.circle_list = [None] * self.display_size

    def reset(self):
        self.clear()
        self.piano.set_scores([0]*12, KeySignature("C", 0, Accidental.SHARP, 0, KeySignatureType.MAJOR))

    def pause_scrollbar(self):
        self.scrollbar.pack(side='bottom', fill='x')

    def resume_scrollbar(self):
        self.scrollbar.pack_forget()
        self.display_recent_notes()





import tkinter as tk
from TuneCoach.python_bridge.pitch_utilities import *
from TuneCoach.gui.constants import *


# set_score uses session.data.key_signature.accidental
class Piano(tk.Canvas):
    def __init__(self, parent, **kwargs):
        super().__init__(parent, kwargs, bd=0, highlightthickness=0)
        self._parent = parent
        self.configure(bg='black')
        self.bind("<Configure>", self.draw)
        self.keys = []
        self.cached_scores = [0] * 12
        self.cached_key = KeySignature("C", 0, Accidental.SHARP, 0, KeySignatureType.MAJOR)
        self.draw(None)

    def draw(self, event):
        self.delete("all")
        self.keys.clear()
        width = self.winfo_width()
        height = self.winfo_height()
        key_gap = 1.2 # TODO: not needed because tk adds a pixel gap automatically?
        white_key_height = height/7
        black_key_height = height/14
        black_key_width = width*0.6
        for i in range(0, 7):
            w = self.create_rectangle(0, (white_key_height)*i + key_gap/2, width, (white_key_height)*(i+1) - key_gap/2, fill='white')
            if i not in {0,4,7}:
                b = self.create_rectangle(0, (white_key_height)*(i) - black_key_height/2, black_key_width, white_key_height*(i)+black_key_height/2, fill='black')
                self.keys.append(b)
            self.keys.append(w)
        self.keys.reverse()
        self.set_scores(self.cached_scores, self.cached_key)
    
    # pass in a list of 12 percentages for C through B
    def set_scores(self, scores, key_signature):
        if len(scores) != 12:
            print("INVALID SCORE LIST")
            return

        self.cached_scores = scores
        self.cached_key = key_signature
        self.delete("text")

        for i in [0, 2, 4, 5, 7, 9, 11]:
            key = self.keys[i]
            coords = self.coords(key)
            x = coords[2] - 30
            y = (coords[3] + coords[1]) / 2
            score = round(scores[i])
            if score >= 70:
                textColor = Colors.green
            elif score >= 50:
                textColor = "#e0b000" # More visible against white background
            elif score == 0:
                textColor = "black"
            else:
                textColor = Colors.red     
            self.create_text(x, y, text=f"{key_signature.get_display_for(i)}: {score}%", fill=textColor, tag="text")
        for i in [1,3,6,8,10]:
            key = self.keys[i]
            coords = self.coords(key)
            x = coords[2] - 30
            y = (coords[3] + coords[1]) / 2
            score = round(scores[i])
            if score >= 70:
                textColor = Colors.green
            elif score >= 50:
                textColor = Colors.yellow
            elif score == 0:
                textColor = "white"
            else:
                textColor = "red" # More visible against black background
            self.create_text(x, y, text=f"{key_signature.get_display_for(i)}: {round(scores[i])}%", fill=textColor, tag="text")


# just a class to create a window letting you know that a new session has been created
import tkinter as tk
import tkinter.ttk as ttk


class NewSessionWindow:
    def __init__(self, mainWindow):
        self.mainWindow = mainWindow
        newSessionWindow = tk.Toplevel(self.mainWindow.master, bg="#F5F6F7")
        newSessionWindow.geometry("300x100")
        myLabel = tk.Label(newSessionWindow, text="New Session Successfully Created.", bg="#F5F6F7")
        myLabel.pack(padx=10, pady=10)
        myButton = ttk.Button(newSessionWindow, text="Ok", command=lambda: newSessionWindow.destroy())
        myButton.pack(padx=10, pady=10)
        newSessionWindow.lift(self.mainWindow.master)



from TuneCoach.gui.MainWindow import *
from TuneCoach.gui.PitchDisplay import *
from tkinter import *
from TuneCoach.python_bridge.AudioManager import *
import time


def main():
    start = time.time()
    root = Tk()
    root.title("TuneCoach")
    our_window = MainWindow(root)
    root.mainloop()


if __name__ == "__main__":
    main()



import tkinter as tk
import tkinter.ttk as ttk
from TuneCoach.gui.constants import *


class TutorialWindow: #TODO: Style Tutorial Window so it's not ugly - Jenny
    def __init__(self, mainWindow):
        tutorial_window = tk.Toplevel(mainWindow.master, bg = "#F4F4F4")
        tutorial_window.title("Tutorial")
        tutorial_window.geometry()
        tutorial_window.minsize(width=590, height=480)

        about_tunecoach_text = "TuneCoach is a convenient PC-based tuning application that records practice sessions and provides detailed feedback \nabout your performance."
        new_to_tunecoach_text = 'New to TuneCoach?'
        get_started_text = 'To get started, press the spacebar to start audio input.\n' \
                           '\nThe Pitch Detector on the right continuously shows the note being played, as well as the intonation.\n' \
                           'Green means that the note is in tune, yellow means that the note is outside set cents threshold, and red means the note is \nvery out of tune.\n' \
                           '\nThe bottom area shows your recently played notes represented on a piano octave. The percentage on each piano key \nreflects how in tune the corresponding note has been played throughout the session. When the session is not recording, \nyou can use the bottom scrollbar to view all the notes you have played.\n\n'\
                           'The Session Diagnostics area on the left shows how your score has changed over time as well as the current settings.\n' \
                           'These can be changed by navigating to Settings->Tuner Settings.'

        keyboard_shortcuts_title = 'Keyboard Shortcuts:'

        tutorial_text_style = ttk.Style()
        tutorial_text_style.configure('TutorialTextLabel.TLabel', padding=10, width=100, anchor=tk.W)
        about_tunecoach_text_label = ttk.Label(tutorial_window, text=about_tunecoach_text, style='TutorialTextLabel.TLabel')
        about_tunecoach_text_label.configure(wraplength = mainWindow.screen_width - 10)
        about_tunecoach_text_label.pack()


        tutorial_title_style = ttk.Style()
        tutorial_title_style.configure('TutorialTitleLabel.TLabel', font="Ubuntu 15", width=73, anchor=tk.W)
        new_to_tunecoach_text_label = ttk.Label(tutorial_window, text=new_to_tunecoach_text, style='TutorialTitleLabel.TLabel')
        new_to_tunecoach_text_label.pack()

        get_started_text_label = ttk.Label(tutorial_window, text=get_started_text, style='TutorialTextLabel.TLabel')
        get_started_text_label.configure(wraplength = mainWindow.screen_width - 80)
        get_started_text_label.pack()

        keyboard_shortcuts_title_label = ttk.Label(tutorial_window, text=keyboard_shortcuts_title, style='TutorialTitleLabel.TLabel', padding=10)
        keyboard_shortcuts_title_label.pack()

        text = ttk.Style()
        text.configure('Text.TLabel', width=100, anchor=tk.W)

        keyboard_shortcut_pairs = [
            ('Space', 'Pause or resume the session.'),
            ('CTRL-N', 'Create a new session.'),
            ('CTRL-S', 'Save the current session.'),
            ('CTRL-SHIFT-S', 'Save the current session as... '),
            ('CTRL-O', 'Open the menu to load an existing session.'),
            ('CTRL-T', 'Open the tuner settings window.'),
            ('F1', 'Open the tutorial window.'),
            ('F2', 'Open the FAQ window.\n')
        ]

        for key, action in keyboard_shortcut_pairs:
            key_label = ttk.Label(tutorial_window, text=key +": "+ action, style='Text.TLabel')
            key_label.pack()

        tutorial_window.lift(mainWindow.master)



class Colors:
	background = '#2d2d2d'
	aux = '#585869'
	text = '#ffffff'
	piano_track = '#bdd0df'
	graph_background = aux # '#85c9ac'
	scroll_trough = 'lightgray' # '#daebe3'
	graph_foreground = 'white'
	score_label = '#60e0a9'
	red = '#ab1302'
	green = '#30bf13'
	yellow = '#ffd324'
	tuner_needle = '#6b3e32' # '#452A23'
	settings_color = '#b6aecf'


# main gui for TuneCoach. Made by the group, Jamm Hostetler, James Eschrich, Joe Gravelle, Jenny Baik, Gavin Gui
from TuneCoach.gui.PitchDisplay import *
from TuneCoach.gui.SessionHistory import *
from TuneCoach.gui.SessionDiagnostics import *
from TuneCoach.gui.TunerSettingsWindow import *
from TuneCoach.gui.FAQWindow import *
from TuneCoach.gui.TutorialWindow import *
from TuneCoach.gui.IntroWindow import *
from TuneCoach.gui.Timer import *

from TuneCoach.gui.MainController import MainController

import tkinter.filedialog
import tkinter.messagebox
import tkinter as tk
from ttkthemes import ThemedStyle


def invalid_path(path):
    # print(path)
    return path == '' or path == () or path == None


# Main GUI
class MainWindow:
    def __init__(self, master):
        style = ThemedStyle(master)
        style.set_theme("yaru")

        self.controller = MainController(self)
        self.master = master

        master.attributes('-fullscreen', True)
        master.state('iconic')
        self.screen_width = master.winfo_screenwidth()
        self.screen_height = master.winfo_screenheight()
        master.attributes('-fullscreen', False)
        master.deiconify()
        master.title("TuneCoach")
        master.geometry(f'{self.screen_width}x{self.screen_height}')
        master.minsize(width=int(self.screen_width * 0.6), height=int(self.screen_height*0.7))
        master.maxsize(width=self.screen_width, height=self.screen_height)

        master.bind('<space>', lambda ev: self.controller.toggle_pause())
        master.bind('<Control-Key-o>', lambda ev: self.controller.load_from())
        master.bind('<Control-Key-t>', lambda ev: TunerSettingsWindow(self))
        master.bind('<Control-Key-n>', lambda ev: self.controller.new_session())
        master.bind('<F1>', lambda ev: TutorialWindow(self))
        master.bind('<F2>', lambda ev: FAQWindow(self))
        master.bind('<F12>', lambda ev: self.controller.save_as())
        master.bind('<Control-Shift-Key-S>', lambda ev: self.controller.save_as())
        if self.controller.save:
            master.bind('<Control-Key-s>', lambda ev: self.controller.save())
        else:
            master.bind('<Control-Key-s>', lambda ev: self.controller.save_as())
        
        self.enable()
        self.create_menubar()
        self.layout_frames(self.screen_width, self.screen_height)
        IntroWindow(self)
    
    def cleanup(self):
        self.disable()
        if not self.controller.cleanup():
            self.enable()
            return # do not close if we're saving and then we cancel

        self.master.destroy() 
        
    # Creating frames to organize the screen.
    def layout_frames(self, screen_width, screen_height):
        frames_style = ttk.Style()
        frames_style.configure('MainFrames.TFrame', background=Colors.background)
        self.bottom_frame = ttk.Frame(self.master, style='MainFrames.TFrame')
        self.left_frame = ttk.Frame(self.master, style='MainFrames.TFrame')
        self.right_frame = ttk.Frame(self.master, style='MainFrames.TFrame')

        # Putting the frames into a grid layout
        self.left_frame.grid(row=0, column=0, sticky="nsew")
        self.right_frame.grid(row=0, column=1, sticky="nsew")
        self.bottom_frame.grid(row=1, column=0, columnspan=2, sticky="nsew")

        # setting up grid weights.
        self.master.grid_rowconfigure(0, weight=1)
        self.master.grid_rowconfigure(1, weight=1, minsize=250)
        self.master.grid_columnconfigure(0, weight=1, uniform="halfwidth")
        self.master.grid_columnconfigure(1, weight=1, uniform="halfwidth")

        # Here we can work on creating the functionality for each frame, ex: tuner, pitch history, information
        self.history = SessionHistory(self, self.bottom_frame)
        self.diagnostics = SessionDiagnostics(self)
        self.pitch_display = PitchDisplay(self, self.controller.threshold)
    
    def perform_save_as(self, newSession = False):
        if newSession:
            path = tk.filedialog.asksaveasfilename(initialdir='./', title="Save session as...", filetypes=[('session files', '*.session')])
        else:
            path = tk.filedialog.asksaveasfilename(initialdir='./', title="Save session as...", filetypes=[('session files', '*.session')])
        if invalid_path(path): # if the user cancels the dialog, don't do anything
            return (None, True) # this tuple means the user canceled

        return (path, False) # we did save

    def perform_load(self):
        path = tk.filedialog.askopenfilename(initialdir='./', title="Select a session to load", filetypes=[('session files', '*.session')])
        if invalid_path(path): # if the user cancels the dialog, don't do anything
            return (None, True) # cancel
        return (path, False)

    # Adding menu options to the top of the screen.
    # returns False ONLY IF THE USER CANCELS
    
    def disable(self):
        self.master.protocol("WM_DELETE_WINDOW", lambda: ...)
        self.controller.pause()

    def enable(self):
        self.master.protocol("WM_DELETE_WINDOW", self.cleanup)
    
    def create_menubar(self):
        def session_menu_item(fn):
            def command_function():
                self.disable()
                fn()
                self.enable()
            return command_function
        
        menubar = tk.Menu(self.master, bg='white')
        self.master.config(menu=menubar)
        file_menu = tk.Menu(menubar)

        # File menubar
        menubar.add_cascade(label="File", menu=file_menu)
        commands = ( 
            ("New Practice Session (CTRL-N)", self.controller.new_session),
            ("Save Current Session (CTRL-S)", self.controller.save),
            ("Save Current Session As... (CTRL-SHIFT-S)", self.controller.save_as),
            ("Load Existing Session (CTRL-O)", self.controller.load_from))
        
        for label, fn in commands:
            file_menu.add_command(label=label, command=session_menu_item(fn), background='white')

        # Settings menubar
        settings_menu = tk.Menu(menubar)
        menubar.add_cascade(label="Settings", menu=settings_menu)
        settings_menu.add_command(label="Tuner Settings (CTRL-T)", command=lambda: TunerSettingsWindow(self), background='white')

        # Help menubar
        help_menu = tk.Menu(menubar)
        menubar.add_cascade(label="Help", menu=help_menu)
        help_menu.add_command(label="Tutorial (F1)", command=lambda: TutorialWindow(self), background='white')
        help_menu.add_command(label="FAQ (F2)", command=lambda: FAQWindow(self), background='white')

   ### METHODS IMPLEMENTED FOR CONTROLLER ### 

    def update_diagnostics(self, data):
        self.diagnostics.update_plot(data)
        self.pitch_display.display_score(data.get_overall())

    def update_history(self, data):
        self.history.update(data)

    def refresh_timer(self, data):
        self.pitch_display.update_data(0, data)

    def update_threshold(self, threshold):
        self.pitch_display.set_threshold(threshold)

    def update_pitch(self, hz, data):
        self.pitch_display.update_data(hz, data)
    
    def update_session_name(self, name):
        self.diagnostics.session_name.configure(text=name)
    
    def clear(self):
        self.history.reset()
        self.diagnostics.clear_plot()
        self.pitch_display.clear()

    def pause(self):
        self.pitch_display.pause()
        self.history.pause_scrollbar()

    def resume(self):
        self.pitch_display.resume()
        self.history.resume_scrollbar()

    def error(self, msg, title="Error!"):
        tk.messagebox.showerror(title, msg)
    
    def success(self, msg, title="Success!"):
        tk.messagebox.showinfo(title, msg)

    def ask_should_save(self):
        return tk.messagebox.askyesnocancel("", "Save current session?")

    def after(self, ms, fn): self.master.after(ms, fn)



import tkinter as tk
from TuneCoach.gui.constants import *
from TuneCoach.gui.RoundedLabel import *
import tkinter.ttk as ttk
from TuneCoach.gui.Graph import *

# OUT data.update_score_history()
class SessionDiagnostics:
    def clear_plot(self):
        self.myGraph.clear_plot()

    def update_plot(self, data):
        self.myGraph.update_plot(data.score_history)
        if data is not None:
            self.score.set_text("Average error: %.1f cents" % data.avg_cents)
            self.thresh_label.set_text("Threshold: ±%d cents" % data.green_thresh)
            self.key_label.set_text("Key Signature: %s" % data.key_signature.name)
            self.range_label.set_text("Range: %s%s to %s%s" % (data.lowest_note, 
                                                              data.lowest_octave, 
                                                              data.highest_note, 
                                                              data.highest_octave))
            data.update_score_history()

    def __init__(self, mainWindow):
        self.mainWindow = mainWindow
        workingFrame = mainWindow.left_frame
        currentSession = mainWindow.controller.session.data
        frames_style = ttk.Style()
        frames_style.configure('DiagnosticsFrame.TFrame', background=Colors.aux)
        title_frame = ttk.Frame(workingFrame, style='DiagnosticsFrame.TFrame')
        graph_frame = ttk.Frame(workingFrame, style='DiagnosticsFrame.TFrame')
        settings_frame = ttk.Frame(workingFrame, style='DiagnosticsFrame.TFrame')

        title_frame.grid(row=0, sticky="nsew", ipady=2.5)
        graph_frame.grid(row=1, sticky="nsew", ipady=2.5)
        settings_frame.grid(row=2, sticky="nsew", ipady=7.5, ipadx=2.5)

        workingFrame.grid_rowconfigure(0, weight=0)
        workingFrame.grid_rowconfigure(1, weight=1)
        workingFrame.grid_rowconfigure(2, weight=0)
        workingFrame.grid_columnconfigure(0, weight=1)

        self.myGraph = Graph(graph_frame)
        self.myGraph.pack(anchor=tk.CENTER, fill=tk.BOTH, expand=True, padx=20)
        title_label_style = ttk.Style()
        title_label_style.configure("TitleLabel.TLabel", font="Ubuntu 20", side=tk.TOP, foreground=Colors.text, background=Colors.aux)
        title_label = ttk.Label(title_frame, text="Current Session", style="TitleLabel.TLabel")
        title_label.pack()

        session_name_label_style = ttk.Style()
        session_name_label_style.configure("SessionName.TLabel", font="Ubuntu 16", foreground="light sky blue", background=Colors.aux)
        self.session_name = ttk.Label(title_frame, text=mainWindow.controller.session.name, style="SessionName.TLabel")
        self.session_name.pack()

        self.score = RoundedLabel(graph_frame, "Average error: 0.0 cents", Colors.score_label, width=225, height=35)
        self.score.pack()

        self.thresh_label = RoundedLabel(settings_frame, "Threshold: ±15 cents", Colors.settings_color, height=45)
        self.thresh_label.grid(row=0, column=0, sticky='ew', padx=5)

        self.key_label = RoundedLabel(settings_frame, "Key Signature: C Major", Colors.settings_color, height=45)
        self.key_label.grid(row=0, column=1, sticky='ew', padx=5)

        self.range_label = RoundedLabel(settings_frame, "Range: C2 to B7", Colors.settings_color, height=45)
        self.range_label.grid(row=0, column=2, sticky='ew', padx=5)

        settings_frame.grid_rowconfigure(0, weight=1)
        for i in range(0, 3):
            settings_frame.grid_columnconfigure(i, weight=1, uniform="settings_labels")
 





from ctypes import *
import threading
from TuneCoach.python_bridge.pitch_utilities import *
from TuneCoach.python_bridge.SessionData import *
from TuneCoach.pitch_detection import TunerStream
from time import sleep


def read(stream, emit):
    while (stream.is_alive()):
        response, success = stream.read()
        if success and response:
            hz = response
            emit(hz)
        sleep(.01)


class AudioManager(TunerStream):
    def __init__(self, emit):
        super().__init__(44100)
        
        self._background_audio = threading.Thread(target=lambda: self.mainloop(), daemon=True)
        self._background_audio.start()
        self._background_reader = threading.Thread(target=lambda: read(self, emit))
        self._background_reader.start()

    def destroy(self):
        self.kill()

    @property
    def paused(self):
        return self.is_paused()


from TuneCoach.python_bridge import SessionData, AudioManager
from TuneCoach.gui.Session import Session, load_session, save_session
from TuneCoach.gui.NewSessionWindow import NewSessionWindow
from collections import deque


class MainController:
    def __init__(self, view):
        self.view = view

        self.threshold = 15
        self.yellow_threshold = 35
        self.paused = True
        self.should_save = False

        self.queue = deque([])
        self.session = Session(SessionData(self.threshold, self.yellow_threshold), None)
        self.audio_manager = AudioManager(lambda hz: self.queue.append(hz))
    
    def cleanup(self):
        if self.should_save:
            v = self.view.ask_should_save()
            if v == None: # cancel
                return False
            elif v == True and not self.save(): # save returns False only if we cancel
                return False
    
        self.audio_manager.destroy()
        return True

    def process_queue(self):
        if not self.paused:
            if len(self.queue) != 0:
                self.should_save = True # we've recieved new data, so we should be ready to save
                top = self.queue.popleft()
                self.session.data.collect_data(top)
                self.update_history()
            self.view.after(20, lambda: self.process_queue())

    def update_diagnostics(self):
        if not self.paused:
            self.view.update_diagnostics(self.session.data)
            self.view.after(500, self.update_diagnostics)

    def update_history(self):
        if self.session.data.has_new_data:
            self.session.data.has_new_data = False # TODO lock
            self.view.update_history(self.session.data)
        # if not self.paused:
        #     self.view.after(20, lambda: self.update_history())

    def update_pitch(self):
        if not self.paused:
            self.view.update_pitch(self.audio_manager.peek(), self.session.data)
        else:
            self.view.update_pitch(0, self.session.data)
        if not self.paused or self.view.pitch_display.needs_update():
            self.view.after(10, self.update_pitch)
    
    def update_tuner_settings(self, cent_threshold, key_signature, from_midi, to_midi):
        self.threshold = cent_threshold
        self.session.data.set_thresholds(cent_threshold, self.yellow_threshold)
        self.session.data.key_signature = key_signature
        self.session.data.midi_range = (from_midi, to_midi)
        self.view.update_threshold(cent_threshold)
        self.refresh_displays()

    def pause(self):
        if self.paused:
            return # don't need to pause again
        
        self.paused = True
        self.view.pause()
        self.audio_manager.pause()
        self.session.data.timer.pause()
        self.flush_queue()
        self.refresh_displays() # TODO: try to avoid events coming in after pause

    def toggle_pause(self, force=False):
        if self.audio_manager.is_paused() and not force:
            self.paused = False
            self.view.resume()
            self.should_save = True
            self.update_session_name()
            self.audio_manager.resume()
            self.process_queue()
            self.update_diagnostics()
            self.update_pitch()
            self.session.data.timer.resume()
        else:
            self.pause()

    def flush_queue(self):
        # add remaining data to session
        while len(self.queue) > 0:
            self.session.data.collect_data(self.queue.popleft())

    def setup_session(self):
        self.reset_everything()

        if self.audio_manager is not None:
            self.audio_manager.kill()
        self.audio_manager = AudioManager(lambda hz: self.queue.append(hz))

        self.refresh_displays()
    
    def reset_everything(self):
        self.pause()
        self.view.clear()

    def update_session_name(self):
        c = "*" if self.should_save else ""
        self.view.update_session_name(self.session.name + c)

    def refresh_displays(self):
        self.update_session_name()
        self.view.update_diagnostics(self.session.data)
        self.view.update_history(self.session.data)
        self.view.refresh_timer(self.session.data)
    
    def save_as(self, newSession = False):
        self.pause()
        path, cancel = self.view.perform_save_as(newSession)
        if cancel:
            return False
        #if not path:
        #    return False
        self.session = self.session.with_path(path)
        # notify session name change
        self.view.update_session_name(self.session.name)
        self._save()
        return True # we did save

    def _save(self):
        save_session(self.session)
        self.view.success(f'Session saved to "{self.session.path}" successfully', title="Session Saved")
        self.should_save = False
        self.view.update_session_name(self.session.name)

    def save(self, newSession = False):
        if not self.should_save:
            if self.session.data.empty:
                self.view.success("Session is empty!")
            else:
                self.view.success("Session already saved!")
            return True
        
        if self.session.path is None:
            return self.save_as(newSession)
        
        self.pause()
        self._save()
        return True

    def new_session(self):
        path = False
        if self.should_save:
            path = self.view.ask_should_save()
        print(path)
        if path is not None:
            if self.should_save and path:
                self.save(True)
            self.view.success("New Session Successfully Created.")
            data = SessionData(self.threshold, self.yellow_threshold)
            self.session = Session(data)
            self.setup_session()

    def load_from(self):
        # if current sesion isn't saved, ask if we should save. If we should
        # try to save. If the user cancels, don't bother to save
        path = False
        if self.should_save:
            path = self.view.ask_should_save()
        if path is not None:
            if self.should_save and path:
                self.save()

            path, cancel = self.view.perform_load()
            if cancel:
                return False
        
            session = load_session(path)

            if session is None:
                self.view.error(f'Session located at "{path}" is invalid!', title="Invalid session")
            else:
                self.session = session
                self.setup_session()
                self.should_save = False
            self.refresh_displays()



import tkinter as tk
import tkinter.ttk as ttk
from TuneCoach.gui.constants import *


class FAQWindow:
    def __init__(self, mainWindow):
        faq_window = tk.Toplevel(mainWindow.master)
        faq_window.title("FAQ")
        faq_window.geometry()
        faq_window.minsize(width=610, height=380)
        faq_window.config(background="#f4f4f4")

        question_answer_list = [
            ('How do I use TuneCoach?', 'Detailed instructions can be found in "Tutorial" under the "Help" dropdown menu.\n'),
            ('Is there any way to change the intonation threshold?', 'Yes! It is located in the "Settings" dropdown menu, under "Tuner Settings" and the "Acceptable Error" slider.\n' \
             'If you set this value to 15, for example, notes within ±15 cents of perfectly in tune will count as correct.\n'),
            ('What do the key signature and range settings do?', 'The current key signature affects how notes are displayed. For example, if D Major is selected (which has C# and F#),\naccidentals will be displayed preferring sharps and C and F will be written with natural signs to indicate they are out of\nthe current key. The range setting affects which notes are detected and can be used to prevent unwanted background \nsound interference. No notes outside the range selected will display in the tuner or history.\n'),
            ('What instruments can I use with TuneCoach?', 'Any instrument can be used with TuneCoach! Even your voice!\n'),
            ('Why does the tuner not work?', 'Check your microphone. Ensure that it is fully plugged in and is recognized as audio input.\n' \
                                             'Other apps and programs may be using the microphone, blocking access to TuneCoach.\n' \
                                             'Other apps must be closed in this case. If it does not help, try restarting TuneCoach or your PC.\n'
                                             'Microphone permissions may be blocked. Check system settings to allow the microphone to be used with TuneCoach.\n'),
            ('Do I need internet to use TuneCoach?', 'No! Once TuneCoach is downloaded and installed, it can be used at any time!\n')
        ]

        faq_question_label_style = ttk.Style()
        faq_question_label_style.configure("FaqQuestionLabel.TLabel", foreground="white", background="grey", font=("Ubuntu 16"), width=70, anchor=tk.W)
        faq_answer_label_style = ttk.Style()
        faq_answer_label_style.configure("FaqAnswerLabel.TLabel", width=100, anchor=tk.W)
        for question, answer in question_answer_list:
            question_label = ttk.Label(faq_window, text= "Q: " + question, style="FaqQuestionLabel.TLabel")
            question_label.pack()
            answer_label = ttk.Label(faq_window, text="\nA: " + answer, style="FaqAnswerLabel.TLabel")
            answer_label.pack()

        faq_window.lift(mainWindow.master)



import tkinter as tk
from TuneCoach.gui.constants import *


class RoundedLabel(tk.Canvas):
    def __init__(self, parent, text, color, transparentColor=Colors.aux, **kwargs):
        self.current_text = text
        self._text = None
        self.color = color
        self.font = tk.font.Font(weight='bold',size=11)
        super().__init__(parent, kwargs, bg=transparentColor, bd=0, highlightthickness=0) # bg=parent["background"]
        self.draw()
        self.bind("<Configure>", self.draw)

    def set_text(self, text):
        self.current_text = text
        self.itemconfig(self._text, text=self.current_text)
        self.draw()

    def draw(self, event=None):
        self.delete("all")
        width = self.winfo_width()
        height = self.winfo_height()
        self.set_font_size(width)
        self.backdrop = self.round_rectangle(0, 0, width, height, radius=30, fill=self.color)
        self._text = self.create_text(width/2, height/2, fill='#424651')

        self.itemconfig(self._text, text=self.current_text, font=self.font)

    def set_font_size(self, width):
        default_size = 11
        self.font.config(size=default_size)
        ratio = default_size/self.font.measure(self.current_text)
        size = min(default_size, int(ratio*width*0.95))
        self.font.config(size=size)

    # thanks https://stackoverflow.com/questions/44099594/how-to-make-a-tkinter-canvas-rectangle-with-rounded-corners
    def round_rectangle(self, x1, y1, x2, y2, radius=25, **kwargs):
        points = [x1+radius, y1,
                x1+radius, y1,
                x2-radius, y1,
                x2-radius, y1,
                x2, y1,
                x2, y1+radius,
                x2, y1+radius,
                x2, y2-radius,
                x2, y2-radius,
                x2, y2,
                x2-radius, y2,
                x2-radius, y2,
                x1+radius, y2,
                x1+radius, y2,
                x1, y2,
                x1, y2-radius,
                x1, y2-radius,
                x1, y1+radius,
                x1, y1+radius,
                x1, y1]
        return self.create_polygon(points, **kwargs, smooth=True)


import tkinter as tk
from TuneCoach.gui.constants import *
from TuneCoach.python_bridge.pitch_utilities import *
import tkinter.ttk as ttk


# Tuner settings window
class TunerSettingsWindow:
    def input_check(self, new_cents, f_note, f_oct, t_note, t_oct, window):
        keytype = KeySignatureType[self.ktype.get().upper()]
        low = string_to_pitch_class(f_note)
        high = string_to_pitch_class(t_note)
        from_midi = pitch_with_octave(low, int(f_oct))
        to_midi = pitch_with_octave(high, int(t_oct))

        if from_midi >= to_midi: 
            self.range_label.pack_forget()
            style = ttk.Style()
            style.configure("Red.TLabel", foreground="red")
            self.range_label = ttk.Label(self.range_frame1, text="Note Range:")
            self.range_label.configure(style="Red.TLabel")
            self.range_label.pack(expand=False, side=tk.LEFT, padx=15)

        else:
            self.update_tuner_settings(new_cents, self.current_key_signature, from_midi, to_midi, window)

    def update_tuner_settings(self, cent_threshold, key_signature, from_midi, to_midi, oldSettingsView):
        self.mainWindow.controller.update_tuner_settings(cent_threshold, key_signature, from_midi, to_midi)
        oldSettingsView.destroy()

    def __init__(self, mainWindow):
        self.mainWindow = mainWindow
        data = mainWindow.controller.session.data
        tuner_settings_window = tk.Toplevel(self.mainWindow.master)
        tuner_settings_window.title("Tuner Settings")
        tuner_settings_window.geometry("400x300")
        tuner_settings_window.minsize(width = 400, height = 300)
        tuner_settings_window.maxsize(width  = 400, height = 300)

        cent_frame = ttk.Frame(tuner_settings_window)
        key_sig_frame = ttk.Frame(tuner_settings_window)
        radio_button_frame = ttk.Frame(tuner_settings_window)
        key_type_frame = ttk.Frame(tuner_settings_window)
        self.range_frame1 = ttk.Frame(tuner_settings_window)
        range_frame2 = ttk.Frame(tuner_settings_window)
        done_frame = ttk.Frame(tuner_settings_window)

        # putting the frames into a grid layout
        cent_frame.grid(row=0, column=0, columnspan=3, sticky="nsew")
        key_sig_frame.grid(row=1, column=0, sticky="nsew")
        radio_button_frame.grid(row=1, column=1, sticky="nsew")
        key_type_frame.grid(row=1, column=2, sticky="nsew")
        self.range_frame1.grid(row=2, column=0, sticky="nsew")
        range_frame2.grid(row=2, column=1, columnspan=2, sticky="nsew")
        self.invalid = False
        done_frame.grid(row=3, column=0, columnspan=3, sticky="nsew")

        # setting up grid weights.
        tuner_settings_window.grid_rowconfigure(0, weight=1)
        tuner_settings_window.grid_rowconfigure(1, weight=1)
        tuner_settings_window.grid_rowconfigure(2, weight=1)
        tuner_settings_window.grid_columnconfigure(0, weight=1, uniform='c')
        tuner_settings_window.grid_columnconfigure(1, weight=1, uniform='c')
        tuner_settings_window.grid_columnconfigure(2, weight=1, uniform='c')

        centsitivity = ttk.Label(cent_frame, text="Acceptable Error (cents):")
        centsitivity.pack(anchor=tk.CENTER, pady=15)

        v = tk.DoubleVar()
        v.set(data.green_thresh)
        cent_scale = tk.Scale(cent_frame, from_=1, to=25, orient=tk.HORIZONTAL, variable=v)
        cent_scale.config(bg="#F4F4F4", fg="black")
        cent_scale.pack(anchor=tk.CENTER)

        sig_label = ttk.Label(key_sig_frame, text="Key Signature:")
        sig_label.pack(expand=False, side=tk.LEFT, padx=15)

        self.range_label = ttk.Label(self.range_frame1, text="Note Range:")
        self.range_label.pack(expand=False, side=tk.LEFT, padx=15)

        # TODO: extract into keysignature and allow for better initialization, create circle-of-fifths-based data structure
        self.major_key_names = ["C", "D♭", "D", "E♭", "E", "F", "F♯", "G", "A♭", "A", "B♭", "B"]
        self.major_accidentals = [Accidental.SHARP, Accidental.FLAT, Accidental.SHARP, Accidental.FLAT,
                                    Accidental.SHARP, Accidental.FLAT, Accidental.SHARP, Accidental.SHARP,
                                        Accidental.FLAT, Accidental.SHARP, Accidental.FLAT, Accidental.SHARP]
        self.major_numbers = [0, 5, 2, 3, 4, 1, 6, 1, 4, 3, 2, 5] # number of sharps/flats in each key
        self.minor_key_names = ["C", "C♯", "D", "E♭", "E", "F", "F♯", "G", "G♯", "A", "B♭", "B"]
        self.minor_accidentals = [Accidental.FLAT, Accidental.SHARP, Accidental.FLAT, Accidental.FLAT,
                                    Accidental.SHARP, Accidental.FLAT, Accidental.SHARP, Accidental.FLAT,
                                        Accidental.SHARP, Accidental.SHARP, Accidental.FLAT, Accidental.SHARP]
        self.minor_numbers = [3, 4, 1, 6, 1, 4, 3, 2, 5, 0, 5, 2]

        self.current_key_signature = data.key_signature
        current_type = data.key_signature.ktype
        self.root = tk.IntVar(value=self.current_key_signature.raw_value)
        self.ktype = tk.StringVar(value=current_type.value)
        self.radio_buttons = []

        # Grid of key signature buttons
        for i in range(0, 12):
            name = ""
            if current_type == KeySignatureType.MINOR:
                name = self.minor_key_names[i]
            else:
                name = self.major_key_names[i]
            button = tk.Radiobutton(radio_button_frame, text=name, indicatoron=0, variable=self.root, value=i, command=self.selection_changed)
            button.grid(row=i//4, column=i%4, sticky='news')
            self.radio_buttons.append(button)

        radio_button_frame.grid_rowconfigure(0, weight=1, uniform="row")
        radio_button_frame.grid_rowconfigure(1, weight=1, uniform="row")
        radio_button_frame.grid_rowconfigure(2, weight=1, uniform="row")
        radio_button_frame.grid_columnconfigure(0, weight=1, uniform="col")
        radio_button_frame.grid_columnconfigure(1, weight=1, uniform="col")
        radio_button_frame.grid_columnconfigure(2, weight=1, uniform="col")
        radio_button_frame.grid_columnconfigure(3, weight=1, uniform="col")

        major_button = tk.Radiobutton(key_type_frame, text="Major", indicatoron=0, width=6, variable=self.ktype, value="Major", command=self.selection_changed)
        major_button.pack(expand=True, side=tk.TOP)
        minor_button = tk.Radiobutton(key_type_frame, text="Minor", indicatoron=0, width=6, variable=self.ktype, value="Minor", command=self.selection_changed)
        minor_button.pack(expand=True, side=tk.BOTTOM)


        # MIDI RANGE SELECTION
        self.from_note = tk.StringVar()
        from_octave = tk.IntVar()
        self.to_note = tk.StringVar()
        to_octave = tk.IntVar()

        self.from_note_menu = ttk.OptionMenu(range_frame2, self.from_note, self.major_key_names[0], *self.major_key_names)
        self.from_note_menu.config(width=2)
        self.from_note_menu.pack(side=tk.LEFT)
        from_octave_menu = ttk.OptionMenu(range_frame2, from_octave, 2, 2, 3, 4, 5, 6, 7)
        from_octave_menu.pack(side=tk.LEFT)

        to_text = ttk.Label(range_frame2, text="to", width=4)
        to_text.configure(anchor="center")
        to_text.pack(side=tk.LEFT)

        self.to_note_menu = ttk.OptionMenu(range_frame2, self.to_note, *self.major_key_names)
        self.to_note_menu.config(width=2)
        self.to_note_menu.pack(side=tk.LEFT)
        to_octave_menu = ttk.OptionMenu(range_frame2, to_octave, 2, 2, 3, 4, 5, 6, 7)
        to_octave_menu.pack(side=tk.LEFT)

        self.from_note.set(data.lowest_note)
        self.to_note.set(data.highest_note)
        from_octave.set(data.lowest_octave)
        to_octave.set(data.highest_octave)

        self.refresh_om()

        done_button = ttk.Button(done_frame, text="Apply", command=lambda: self.input_check(cent_scale.get(), self.from_note.get(), from_octave.get(), self.to_note.get(), to_octave.get(), tuner_settings_window))
        done_button.pack()
        tuner_settings_window.lift(self.mainWindow.master)

    def selection_changed(self, redraw=True):
        index = self.root.get()
        keytype = KeySignatureType[self.ktype.get().upper()]
        num_accidentals = 0
        if keytype == KeySignatureType.MINOR:
            accidental = self.minor_accidentals[index]
            name = self.minor_key_names[index]
            num_accidentals = self.minor_numbers[index]
        else:
            accidental = self.major_accidentals[index]
            name = self.major_key_names[index]
            num_accidentals = self.major_numbers[index]
            
        self.current_key_signature = KeySignature(name, index, accidental, num_accidentals, keytype)

        if redraw:
            self.refresh_om()
            if self.ktype.get() == "Minor":
                names = self.minor_key_names
            else:
                names = self.major_key_names
            for i in range(0, 12):
                self.radio_buttons[i].config(text=names[i])

    def refresh_om(self):
        low = string_to_pitch_class(self.from_note.get())
        high = string_to_pitch_class(self.to_note.get())

        self.from_note_menu['menu'].delete(0, 'end')
        self.to_note_menu['menu'].delete(0, 'end')

        notes = [self.current_key_signature.get_display_for(n) for n in range(0,12)]
        start = self.current_key_signature.raw_value
        notes = notes[start:] + notes[:start] # rotate so current root is first

        for note in notes:
            self.from_note_menu['menu'].add_command(label=note, command=tk._setit(self.from_note, note))
            self.to_note_menu['menu'].add_command(label=note, command=tk._setit(self.to_note, note))

        self.from_note.set(self.current_key_signature.get_display_for(low))
        self.to_note.set(self.current_key_signature.get_display_for(high))




from tkinter import *
from tkinter.font import Font
import tkinter.ttk as ttk
from TuneCoach.python_bridge.pitch_utilities import *
from math import sin, cos, radians
import time
from TuneCoach.gui.IndicatorLight import *
from TuneCoach.gui.constants import *
from TuneCoach.gui.RoundedLabel import *


class PitchDisplay:
    def __init__(self, mainWindow, threshold=10):
        self.frame = mainWindow.right_frame
        self.mainWindow = mainWindow

        self.threshold = threshold

        self.font = Font(size=20)
        self.pitchOffset = self.font.metrics('linespace')*0.75

        self._pitchValue = '---'  # default display
        self._centsValue = -50
        self._hertzValue = 0
        self._octaveValue = ''

        self._span = 75  # Size of tuner arc in degrees, starting at vertical

        self.canvas = Canvas(self.frame, bg=Colors.background, bd=0, highlightthickness=0)
        self.canvas.pack(fill=BOTH, expand=True)
        self.canvas.bind("<Configure>", self.configure)

        self.top_frame = Frame(self.canvas, height=35, bg=Colors.background, bd=0, highlightthickness=0)
        #self.top_frame.pack_propagate(0)
        self.top_frame.pack(side='top', fill=tk.X, anchor=tk.N)

        self.rec_frame = Frame(self.top_frame, width=80, height=35, bg=Colors.background, bd=0, highlightthickness=0)
        self.rec_frame.pack_propagate(0)
        self.rec_frame.pack(anchor='w', side=tk.LEFT)

        self.light = IndicatorLight(self.rec_frame, 35)
        self.light.pack(anchor='w', side='left')

        self.time_label = Label(self.rec_frame, text='00:00', anchor='e', justify=RIGHT, fg=Colors.text, bg=Colors.background)
        self.time_label.pack(side='right')

        self.score_label = RoundedLabel(self.top_frame, "Score: 0%", Colors.score_label, Colors.background, height=35, width=110)
        self.score_label.pack(side='right')
        #self.score_label.set_text("adfasdf")

        self.showsHertz = BooleanVar()

        style = ttk.Style()
        style.configure("Pitch.TCheckbutton", background=Colors.background, foreground=Colors.text)
        style.map('Pitch.TCheckbutton',

        foreground=[('active', Colors.text)],
        background=[('pressed', '!focus', '#232323'),
                    ('active', Colors.aux)])

        c = ttk.Checkbutton(self.canvas, text="Show Hertz", variable=self.showsHertz, takefocus=False, command=self.display_default_gui, style="Pitch.TCheckbutton")
        c.pack(anchor='e', side='bottom')

        self._last_time = 0
        self._clearing = False

        self.display_default_gui()

    def pause(self):
        self.light.stop()
        self.canvas.itemconfig(self.help_text, text='Press \'space\' to accept audio input')
        self._clearing = True

    def resume(self):
        self.light.start_flashing()
        self.canvas.itemconfig(self.help_text, text='Press \'space\' to pause audio input')

    def cents_to_angle(self, cents):
        return cents/50 * self._span

    def configure(self, event):
        self.display_default_gui()

    def display_score(self, score):
        self.score_label.set_text(f"Score: {round(score)}%")

    def display_current_gui(self):
        pitch_and_octave = self._pitchValue + self._octaveValue
        self.canvas.itemconfig(self.current_pitch_display, text=pitch_and_octave)
        if self.showsHertz.get():
            self.canvas.itemconfig(self.hertzDisplay, text=self._hertzValue)
        self.update_line(self._centsValue)
        if not self._clearing and abs(self._centsValue) <= self.threshold:
            self.canvas.itemconfig(self.green_arc, fill=Colors.green)
        else:
            self.canvas.itemconfig(self.green_arc, fill="#ccffbf")

    def display_default_gui(self):
        self.canvas.delete("all")
        self.width = self.frame.winfo_width()
        self.height = self.frame.winfo_height()
        min_dimension = min(self.width, self.height)
        self.radius = 0.4 * min_dimension
        self.centerX = self.width/2
        self.centerY = self.height/2 + 15

        self.current_pitch_display = self.canvas.create_text(self.centerX, self.centerY + self.pitchOffset, font=self.font, text='---', fill=Colors.text)
        if self.showsHertz.get():
            self.hertzDisplay = self.canvas.create_text(self.centerX, self.centerY + 2*self.pitchOffset, font="Ubuntu 14", text='', fill=Colors.text)

        self.help_text = self.canvas.create_text(self.width/2, self.height-35, text='Press \'space\' to accept audio input', fill=Colors.text)

        x0 = self.centerX - self.radius
        y0 = self.centerY - self.radius
        x1 = self.centerX + self.radius
        y1 = self.centerY + self.radius

        # rect = self.canvas.create_rectangle(x0, y0, x1, y1)

        rStart = 90 - self._span
        rSpan = 2 * self._span
        yStart = 90 - self.cents_to_angle(self.mainWindow.controller.yellow_threshold)
        ySpan = 2 * (90 - yStart)
        gStart = 90 - self.cents_to_angle(self.threshold)
        gSpan = 2 * self.cents_to_angle(self.threshold)

        self.red_arc = self.canvas.create_arc(x0, y0, x1, y1)
        self.canvas.itemconfig(self.red_arc, start=rStart, fill="#ffbfbf", extent=rSpan, outline='')

        self.yellow_arc = self.canvas.create_arc(x0, y0, x1, y1)
        self.canvas.itemconfig(self.yellow_arc,  start=yStart, fill="#fffeb0", extent=ySpan, outline='')

        self.green_arc = self.canvas.create_arc(x0, y0, x1, y1)
        self.canvas.itemconfig(self.green_arc, start=gStart, fill="#ccffbf", extent=gSpan, outline='')

        self.line = self.canvas.create_line(0, 0, 0, 0, fill=Colors.tuner_needle, width=4, 
                                            arrow=FIRST, arrowshape=(self.radius, 10, 5))
        self.update_line(-50)

    def update_line(self, cents):
        deg = self.cents_to_angle(cents)
        theta = radians(deg)
        dx = self.radius * sin(theta)
        dy = self.radius * cos(theta)
        self.canvas.coords(self.line, self.centerX, self.centerY, self.centerX + dx, self.centerY - dy)

    def set_threshold(self, thresh):
        self.threshold = thresh
        self.display_default_gui()

    def update_pitch(self, value):  # event as parameter
        self._pitchValue = value

    def update_hertz(self, value):
        self._hertzValue = value

    def update_cents(self, value):
        self._centsValue = value

    def update_octave(self, value):
        self._octaveValue = value

    def set_time(self, total_seconds):
        minutes = total_seconds // 60
        seconds = total_seconds % 60
        display_string = '{:02}:{:02}'.format(minutes, seconds)
        self.time_label.config(text=display_string)

    def update_data(self, hz, data): #TODO: remove data parameter?
        if hz != 0:
            self._clearing = False
            midi = hz_to_midi(hz)
            if data.midi_range[0] <= midi <= data.midi_range[1]:
                pitch_class = midi_to_pitch_class(midi)
                desired_hz = closest_in_tune_frequency(hz)
                cent = cents(desired_hz, hz)
                name = data.key_signature.get_display_for(pitch_class)
                self.update_cents(cent)
                self.update_hertz(f"{round(hz)} Hz")
                self.update_octave(f"{get_octave(midi)}")
                self.update_pitch(name)
                self.display_current_gui()
                self._last_time = time.time()
        else:
            self.clear()

        self.set_time(data.timer.get()) # TODO move timer

    def clear(self):
        self._clearing = True
        if self._centsValue != -50 and time.time() - self._last_time > 1.5:
            self.update_cents(max(-50, self._centsValue - 3))
            self.update_pitch('---')
            self.update_hertz('')
            self.update_octave('')
            self.display_current_gui()
        elif self._centsValue == -50: # clearing animation has finished
            self._clearing = False

    def needs_update(self):
        return self._clearing


import collections
from TuneCoach.python_bridge.pitch_utilities import *
from TuneCoach.gui.MainWindow import *
from TuneCoach.gui.Timer import *
import math
import datetime
import pickle
import os
from pathlib import Path
from collections import namedtuple

Note = collections.namedtuple('Note', 'hz cents pitch_class octave')


def save_to_file(session, path):
    with open(path, "wb") as file:
        pickle.dump(session, file)


def load_from_file(path):
    try:
        with open(path, "rb") as file:
            session = pickle.load(file)
            return session
    except Exception as e:
        # print(e)
        return None


class SessionData:
    def __init__(self, green_thresh, yellow_thresh):
        self._in_tune_count = [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
        self._pitch_count = [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
        self.note_history = []
        self._cents = 0.0
        self._overall = 0
        self._overall_count = 0
        self.green_thresh = green_thresh
        self.yellow_thresh = yellow_thresh
        self._timestamp = datetime.date.today()
        self.has_new_data = False

        self.key_signature = KeySignature("C", 0, Accidental.SHARP, 0, KeySignatureType.MAJOR)

        self.midi_range = (36, 107)

        self.timer = Timer()
        self.timer.start()
        self.timer.pause()

        self._score_history = []
        self.score_history_size = 15

    def get_overall(self):
        if self._overall_count == 0:
            return 0
        else:
            return (100.0 * self._overall) / self._overall_count

    @property
    def avg_cents(self):
        if self._overall_count == 0:
            return 0.0
        else:
            return self._cents / self._overall_count

    @property
    def score_history(self): return self._score_history

    @property
    def empty(self):
        return self._overall_count == 0

    @property
    def lowest_octave(self):
        return get_octave(self.midi_range[0])

    @property
    def highest_octave(self):
        return get_octave(self.midi_range[1])
    
    @property
    def lowest_note(self):
        return self.key_signature.get_display_for(self.midi_range[0] % 12)

    @property
    def highest_note(self):
        return self.key_signature.get_display_for(self.midi_range[1] % 12)
    
    def update_score_history(self):
        new_score = self.get_overall()
        self._score_history.append(new_score)
        if len(self._score_history) > self.score_history_size:
                self._score_history.pop(0)

    def set_thresholds(self, green_thresh, yellow_thresh):
        self.green_thresh = green_thresh
        self.yellow_thresh = yellow_thresh

    # Takes in frequency and calculates and stores all data
    def collect_data(self, hz):
        midi = hz_to_midi(hz)
        if not (self.midi_range[0] <= midi <= self.midi_range[1]):
            return
        
        self.has_new_data = True
        index = midi_to_pitch_class(midi)
        desired_hz = closest_in_tune_frequency(hz)
        cent = cents(desired_hz, hz)
        octave = get_octave(midi)
        self.note_history.append(Note(hz=hz, cents=cent, pitch_class=index, octave=octave))

        # Gets counts of everything to calculate accuracy
        if abs(cent) <= self.green_thresh:
            self._in_tune_count[index] += 1
            self._overall += 1
        elif abs(cent) <= self.yellow_thresh:
            weight = 0.5 # TODO: scale by cent value?
            self._in_tune_count[index] += weight
            self._overall += weight

        self._pitch_count[index] += 1
        self._overall_count += 1
        self._cents += abs(cent)




import tkinter.ttk as ttk
import tkinter as tk
from TuneCoach.gui.constants import Colors


class IntroWindow:
    @staticmethod
    def intro_load_session(mainWindow, oldFrame):
        oldFrame.destroy()
        mainWindow.disable()
        mainWindow.controller.load_from()
        mainWindow.enable()

    @staticmethod
    def intro_save_session_as(mainWindow, oldFrame):
        oldFrame.destroy()
        mainWindow.disable()
        mainWindow.controller.save_as()
        mainWindow.enable()

    def __init__(self, mainWindow):
        self.master = mainWindow.master
        introWindow = tk.Toplevel(self.master)
        introWindow.geometry("520x300")
        introWindow.minsize(width = 520, height = 350)
        introWindow.maxsize(width = 520, height = 350)
        intro_window_style = ttk.Style()
        intro_window_style.configure("Intro.TFrame")
        topFrame = ttk.Frame(introWindow, style="Intro.TFrame")
        middleFrame = ttk.Frame(introWindow, style="Intro.TFrame")
        bottomFrame = ttk.Frame(introWindow, style="Intro.TFrame")

        topFrame.grid(row=0, column=1, sticky="nsew")
        middleFrame.grid(row=1, column=1, sticky="nsew")
        bottomFrame.grid(row=2, column=1, sticky="nsew")

        introWindow.grid_rowconfigure(0, weight=1)
        introWindow.grid_rowconfigure(1, weight=1)
        introWindow.grid_rowconfigure(2, weight=1)

        intro_label_style = ttk.Style()
        intro_label_style.configure("IntroTitle.TLabel", font="Ubuntu 24", padding=15)
        introLabel = ttk.Label(topFrame, text="Welcome to Tune Coach!", anchor=tk.CENTER, style="IntroTitle.TLabel")
        introLabel.pack(side=tk.TOP)

        text = "New to TuneCoach? Checkout Help -> Tutorial.\n\n" \
               "Click 'Load Session' if you would like to load a previous session from file. \n\n" \
               "If you don't load an existing session, a temporary session will be used. You can choose to store this session under a name by clicking the 'Save Temporary Session As...' button.\n"

        explain_label_style = ttk.Style()
        explain_label_style.configure("Intro.TLabel", wraplength=375)
        explainLabel = ttk.Label(middleFrame, text=text, style="Intro.TLabel")
        explainLabel.pack()
        loadButton = ttk.Button(bottomFrame, text="Load Session",
                               command=lambda: self.intro_load_session(mainWindow, introWindow))
        loadButton.pack(side=tk.RIGHT, padx=50, pady=20)
        newSession = ttk.Button(bottomFrame, text="Save Temporary Session As...",
                               command=lambda: self.intro_save_session_as(mainWindow, introWindow))
        newSession.pack(side=tk.LEFT, padx=50, pady=20)

        introWindow.lift(self.master)



from datetime import datetime
import time

# Big thanks to https://stackoverflow.com/questions/60026296/how-to-make-a-pausable-timer-in-python


class Timer:
    def __init__(self):
        self.started = None
        self.paused = None
        self.is_paused = False

    def start(self):
        self.started = datetime.now()

    def pause(self):
        self.paused = datetime.now()
        self.is_paused = True

    def resume(self):
        pause_time = datetime.now() - self.paused
        self.started += pause_time
        self.is_paused = False

    def get(self):
        if self.is_paused:
            return int((self.paused - self.started).total_seconds())
        else:
            return int((datetime.now() - self.started).total_seconds())

    def clear(self):
        self.started = None
        self.paused = None
        self.is_paused = False

    def has_started(self):
        if self.started is None:
            return False
        else:
            return True



import math
from enum import Enum
from dataclasses import dataclass


class Accidental(Enum):
    FLAT = "♭"
    SHARP = "♯"
    NATURAL = "♮"


class KeySignatureType(Enum):
    MAJOR = "Major"
    MINOR = "Minor"

class Notes:
    FLAT_NOTES = ["C", "D♭", "D", "E♭", "E", "F", "G♭", "G", "A♭", "A", "B♭", "B"]
    SHARP_NOTES = ["C", "C♯", "D", "D♯", "E", "F", "F♯", "G", "G♯", "A", "A♯", "B"]
    NOTE_DICT = {
        "C": 0,
        "C♮": 0,
        "C♯": 1,
        "D♭": 1,
        "D": 2,
        "D♮": 2,
        "D♯": 2, 
        "E♭": 3, 
        "E": 4,
        "E♮": 4, 
        "F": 5,
        "F♮": 5,
        "F♯": 6, 
        "G♭": 6, 
        "G": 7,
        "G♮": 7,
        "G♯": 8, 
        "A♭": 8, 
        "A": 9,
        "A♮": 9, 
        "A♯": 10,
        "B♭": 10,
        "B": 11,
        "B♮": 11,
        "C♭": 11,
    }

@dataclass
class KeySignature:
    root: str
    raw_value: int
    accidental: Accidental
    num_accidentals: int
    ktype: KeySignatureType
    
    # raw_value should be in [0,11]
    def get_display_for(self, raw_value):
        # TODO: don't create these on every function call
        flat_order = [11, 4, 9, 2, 7, 0, 5] # BEADGCF
        sharp_order = list(reversed(flat_order)) # FCGDAEB
        notes = {0: "C", 2: "D", 4: "E", 5: "F", 7: "G", 9: "A", 11: "B"}

        relevant_accidentals = []
        if self.accidental == Accidental.SHARP:
            relevant_accidentals = sharp_order[:self.num_accidentals]
        elif self.accidental == Accidental.FLAT:
            relevant_accidentals = flat_order[:self.num_accidentals]

        if raw_value in relevant_accidentals:
            return notes[raw_value] + Accidental.NATURAL.value
        else:
            return pitch_class_to_name(raw_value, self.accidental)

    @property
    def name(self):
        root_str = self.get_display_for(self.raw_value)
        return f"{root_str} {self.ktype.value}"


def hz_to_midi(hz):
    return int(round(12.0 * math.log(hz/440.0, 2) + 69.0))


def midi_to_hz(midi):
    return 440.0 * 2.0**((midi - 69)/12.0)


def midi_to_pitch_class(midi):
    return int(midi % 12)


def get_octave(midi):
    return int(midi/12) - 1


def pitch_with_octave(pitch_class, octave):
    return 12 * (octave + 1) + pitch_class


def pitch_class_to_name(pitch, acc):
    if acc == Accidental.FLAT:
        return Notes.FLAT_NOTES[pitch]
    elif acc == Accidental.SHARP:
        return Notes.SHARP_NOTES[pitch]


def cents(target, actual):
    return 1200.0 * math.log(actual/target, 2)


def closest_in_tune_frequency(hz):
    return midi_to_hz(hz_to_midi(hz))


def string_to_pitch_class(string):
    return Notes.NOTE_DICT.get(string, 0) #TODO: more robust error handling
